#include <DatSubsystems/DatPatcher.h>
#include <SubfileData.h>
#include <DatFile.h>
#include <yaml-cpp/yaml.h>

namespace LOTRO_DAT {

    DatPatcher::DatPatcher(DatFile *datFilePtr) : dat(datFilePtr) {
    }

    /*!
     * \Author Gi1dor
     * \date 07.07.2018
     * Обновление файла. Записывает данные файла в dat контейнер, используя информацию в SubfileData. Если ранее файл не существовал - он не будет создан (вернёт ошибку).
     * \warning В процессе применения локаль будет автоматически переключена в положение PATCHED
     * \param[in] data Новые данные файла
     */

    DatOperationResult<> DatPatcher::PatchFile(const SubfileData &data) {
        auto file_id = data.options["fid"].as<long long>();

        auto getfile_operation = dat->getFileSystem().GetFile(file_id);
        if (getfile_operation.result == ERROR)
            return DatOperationResult<>(ERROR,
                                        "PATCHSUBFILEDATA: Unable to find file with id " + std::to_string(file_id));

        auto& file = getfile_operation.value;

        // If file has inactive category, then we should set it to patched state in order to commit patch and
        // then in ApplyFilePatch() function, if new category is still inactive, return dictionary to its original state;

        if (dat->getLocaleManager().CategoryIsInactive(file->category) != 0) {
            auto operation = dat->getLocaleManager().GetLocaleFile(file->file_id(), DatLocaleManager::PATCHED);
            if (operation.result == SUCCESS)
                dat->getFileSystem().UpdateFileInfo(operation.value);
        }

        if (data.options["cat"].IsDefined())
            file->category = data.options["cat"].as<long long>();
        else
            file->category = 1;

        auto getdata_operation = dat->getFileSystem().GetFileData(file_id, 0);
        if (getdata_operation.result == ERROR)
            return DatOperationResult<>(ERROR,
                                        "PATCHSUBFILEDATA: can't get file data for id = " + std::to_string(file_id));
        auto& file_data = getdata_operation.value;

        BinaryData patch_data = file->MakeForImport(file_data, data);

        auto result = ApplyFilePatch(file, patch_data);
        if (result.result == ERROR)
            return DatOperationResult<>(ERROR,
                                        "PATCHSUBFILEDATA: applyfilepatch failed for id = " + std::to_string(file_id));

        return DatOperationResult<>(SUCCESS);
    }


    /*!
     * \Author Gi1dor
     * \date 07.07.2018
     * Обновление всех файлов в Database.
     * \warning В процессе применения локаль будет автоматически переключена в положение PATCHED
     * \param[in] db Указатель на базу данных. База должна быть проинициализирована
     */

    DatOperationResult<int> DatPatcher::PatchAllDatabase(Database *db) {
        if (!db)
            return DatOperationResult<int>(0, ERROR, "PATCHALLDATABASE: db is nullptr");

        SubfileData data;
        data = db->GetNextFile();

        int successfully_patched = 0;

        while (!data.Empty()) {
            auto operation = PatchFile(data);
            if (operation.result == SUCCESS)
                successfully_patched++;

            data = db->GetNextFile();
        }

        LOG(INFO) << "Successfully patched whole database";
        return DatOperationResult<int>(successfully_patched, SUCCESS);
    }


    /*!
     * \Author Gi1dor
     * \date 07.07.2018
     * Функция, вызываемая функцией PatchFile, отвечающая за корректную запись файла в dat контейнер и обновление
     * информации о нём в файловой системе
     * \param[in] file Указатель на объект файла в файловой системе (полученный через DatFileSystem::GetFileInfo)
     * \param[in] data Бинарные данные собранного нового файла, готовые к записи в dat
     */

    DatOperationResult<> DatPatcher::ApplyFilePatch(std::shared_ptr<SubFile> file, BinaryData &data) {
        long long file_id = file->file_id();

        if (dat->getLocaleManager().GetCurrentLocale() != DatLocaleManager::PATCHED) {
            LOG(INFO) << "Changing locale to PATCHED(RU) in order to patch file";
            dat->getLocaleManager().SetLocale(DatLocaleManager::PATCHED);
        }

        LOG(INFO) << "Patching file with id = " << file_id;

        if (dat->getLocaleManager().GetLocaleFile(file_id, DatLocaleManager::ORIGINAL).result == ERROR)
            dat->getLocaleManager().UpdateLocaleFile(DatLocaleManager::ORIGINAL, SubFile(*file));


        SubFile new_file = SubFile(*file);

        if (dat->getLocaleManager().GetLocaleFile(file_id, DatLocaleManager::PATCHED).result == ERROR
            || data.size() > file->block_size()) {

            new_file.file_offset_ = dat->getIO().file_size;
            new_file.block_size_ = std::max(data.size(), 256u);
            dat->getIO().file_size += new_file.block_size_ + 8;
        }
        new_file.file_size_ = data.size() - 8;


        data.Append(BinaryData::FromNumber<4>(0), 0); // set additional fragments count to zero

        if (file_id != data.ToNumber<4>(8))
            LOG(WARNING) << "Created data's file_id " << file_id << "doesn't match to original: " << data.ToNumber<4>(8);

        auto operation = dat->getIO().WriteData(data, data.size(), new_file.file_offset());
        if (operation.result == ERROR)
            return DatOperationResult<>(ERROR, "APPLYPATCHFILE: Unable to write data for file with id " + std::to_string(file_id));


        dat->getFileSystem().UpdateFileInfo(new_file);
        dat->getLocaleManager().UpdateLocaleFile(DatLocaleManager::PATCHED, new_file);


        // If file category is inactive, then return file header data in dictionary to original state
        if (dat->getLocaleManager().CategoryIsInactive(new_file.category))
            dat->getFileSystem().UpdateFileInfo(dat->getLocaleManager().GetLocaleFile(file_id, DatLocaleManager::ORIGINAL).value);

        dat->getLocaleManager().UpdateCategory(file_id, new_file.category);
        LOG(INFO) << "Successfully patched file with id = " << file_id;

        return DatOperationResult<>(SUCCESS);
    }
};